import unittest

from inspect import getmodule, currentframe
from typing import Callable

from unittest_specs.fun_test_spec import describe, it, expect


class FunTestDSL(unittest.TestCase):
    def test_it_should_return_correct_name(self):
        actual_name, _ = it("a test name", lambda: 1)
        self.assertEqual(actual_name, "test_a_test_name")

    def test_it_should_return_passed_function(self):
        _, actual_def = it("", lambda: 1)
        self.assertEqual(actual_def(), 1)

    def test_describe_should_register_class(self):
        describe("Test Class")
        module = getmodule(currentframe())
        self.assertEqual(module.__dict__["TestClass"].__name__, "TestClass")

    def test_exception_interception(self):
        def exception_raiser(_):
            raise Exception()

        _, wrapped_interception_def = it("", exception_raiser, intercept=Exception)
        wrapped_interception_def(self)


class FunAsserterSpec(unittest.TestCase):
    def setUp(self) -> None:
        self.five_asserter = expect(5)
        self.list_asserter = expect([1, 2, 3])
        self.dict_asserter = expect({1: 3, 2: 2, 3: 1})
        self.set_asserter = expect({1, 2, 3})
        self.none_asserter = expect(None)
        self.true_asserter = expect(True)
        self.false_asserter = expect(False)

    def test_should_create_asserter(self):
        self.assertIsInstance(expect(''), unittest.TestCase)

    def test_asserter_should_return_callable(self):
        self.assertIsInstance(expect('').to_be(''), Callable)

    def test_asserter_should_detect_equality_correctly(self):
        self.five_asserter.to_be(5)(self)

    def test_asserter_should_detect_inequality_correctly(self):
        self.five_asserter.to_not_be(2)(self)

    def test_asserter_should_detect_type_correctly(self):
        self.five_asserter.to_be_of_type(int)(self)

    def test_asserter_should_detect_list_equality_correctly(self):
        self.list_asserter.to_equal_list([1, 2, 3])(self)

    def test_asserter_should_detect_list_contains_correctly(self):
        self.list_asserter.to_contain(2)(self)

    def test_asserter_should_detect_list_contains_all_correctly(self):
        self.list_asserter.to_contain_all([3, 1, 2])(self)

    def test_asserter_should_detect_true_correctly(self):
        self.true_asserter.to_be_true()(self)

    def test_asserter_should_detect_false_correctly(self):
        self.false_asserter.to_be_false()(self)

    def test_asserter_should_detect_none_correctly(self):
        self.none_asserter.to_be_none()(self)

    def test_asserter_should_detect_not_none_correctly(self):
        self.five_asserter.to_not_be_none()(self)

    def test_asserter_should_detect_list_type_correctly(self):
        self.list_asserter.to_be_a_list()(self)

    def test_asserter_should_detect_dict_type_correctly(self):
        self.dict_asserter.to_be_a_dict()(self)

    def test_asserter_should_detect_set_type_correctly(self):
        self.set_asserter.to_be_a_set()(self)

    def test_asserter_should_detect_list_length_correctly(self):
        self.list_asserter.to_be_of_length(3)(self)



import unittest
from typing import Any


class SimpleFlatSpec(unittest.TestCase):
    @staticmethod
    def expect(actual_value: Any):
        """
        Set up an 'actual' value to be compared against an expectation/assertion.

        :param actual_value: the value computed by the SUT to be compared against an expectation
        :return: an Asserter to perform an assertion on
        """
        return Asserter(actual_value)

    @staticmethod
    def parameterize(params):
        """
        Injects provided test data sets into the decorated function with every data set being provided as a set of
        parameters. For each set, a separate sub-test is performed using unittest's subTest.

        :param params: a list of tuples containing the test data
        """

        def decorator_function(function):
            def parameter_handler(self):
                for param in params:
                    if isinstance(param, tuple):
                        test_params = [_p for _p in param]
                    else:
                        test_params = [param]

                    with self.subTest(i=test_params[0]):
                        function(self, *test_params)

            return parameter_handler

        return decorator_function

    @staticmethod
    def intercept(expected_exception):
        """
        Intercepts an expected Exception object occurring in the decorated test function. If no Exception
        of the specified type is raised, the test fails.

        :param expected_exception: the Exception type that is expected to occur
        """

        def decorator_function(function):
            def nested_execution_handler(self, *args, **kwargs):
                interceptor = unittest.TestCase()

                with interceptor.assertRaises(expected_exception=expected_exception):
                    function(self, *args, **kwargs)

            return nested_execution_handler

        return decorator_function


class Asserter(unittest.TestCase):
    def __init__(self, actual_value: Any):
        super().__init__()
        self._actual_value = actual_value

    def to_be(self, expected_value):
        self.assertEqual(self._actual_value, expected_value)

    def to_not_be(self, expected_value):
        self.assertNotEqual(self._actual_value, expected_value)

    def to_be_of_type(self, expected_type):
        self.assertIsInstance(self._actual_value, expected_type)

    def to_equal_list(self, expected_list):
        self.assertListEqual(self._actual_value, expected_list)

    def to_contain(self, expected_element):
        self.assertTrue(expected_element in self._actual_value)

    def to_contain_all(self, expected_elements):
        is_contained = True
        for element in expected_elements:
            if element not in self._actual_value:
                is_contained = False
                break
        self.assertTrue(is_contained)

    def to_be_true(self):
        self.assertTrue(self._actual_value)

    def to_be_false(self):
        self.assertFalse(self._actual_value)

    def to_be_none(self):
        self.assertIsNone(self._actual_value)

    def to_not_be_none(self):
        self.assertIsNotNone(self._actual_value)

    def to_be_a_list(self):
        self.to_be_of_type(list)

    def to_be_a_dict(self):
        self.to_be_of_type(dict)

    def to_be_a_set(self):
        self.to_be_of_type(set)

    def to_raise(self, expected_exception, *args, **kwargs):
        with self.assertRaises(expected_exception=expected_exception):
            self._actual_value(*args, **kwargs)

    def to_be_of_length(self, expected_length: int):
        """
        Asserts whether the previously provided value's length matches expected_length. The length is determined by
        using the BIF len().
        
        :param expected_length: the expected value of len(actual_value)
        """
        self.assertEqual(len(self._actual_value), expected_length)



import unittest
from inspect import stack, getmodule
from typing import Callable, Tuple, Type, Any


def describe(description: str, *test_config) -> None:
    """
    Constructs a test suite containing zero or more test cases.

    :param description: intended for documentation of this describe() block; this description is transformed into
    the type name of the generated unittest.TestCase subclass
    :param test_config: zero or more test case defined by it() blocks
    """
    class_name = description.title().replace(" ", "")

    test_class = type(class_name, (unittest.TestCase,),
                      {test_name: test_function for test_name, test_function in test_config})

    module = getmodule(stack()[1][0])
    module.__dict__[class_name] = test_class


def it(description: str, test_def: Callable, intercept: Type[Exception] = None) -> Tuple[str, Callable]:
    """
    Constructs a test case consisting of a description and an assertion line.

    :param description: intended for documentation of this test case; the description is also transformed into the
    test method name
    :param test_def: assertion line passed to describe() block for
    :param intercept: Intercepts an expected Exception object occurring in this it() declaration. If no Exception
    of the specified type is raised, the test fails.
    :return: a tuple composed of the test method name and the assertion line; this is only intended to be
    used by describe()
    """

    if intercept and issubclass(intercept, Exception):
        def intercept_block(_):
            with unittest.TestCase().assertRaises(expected_exception=intercept):
                test_def()

        return f"test_{description.replace(' ', '_')}", intercept_block

    return f"test_{description.replace(' ', '_')}", test_def


def expect(actual_value: Any):
    """
    Sets up an actual value to be compared against an expectation

    :param actual_value: the value computed by the function(ality) to be tested
    :return: an Asserter object providing different ways to define the expected result
    """
    class Asserter(unittest.TestCase):
        def to_be(self, expected_value) -> Callable:
            def run_test(_=None):
                self.assertEqual(actual_value, expected_value)

            return run_test

        def to_not_be(self, expected_value) -> Callable:
            def run_test(_=None):
                self.assertNotEqual(actual_value, expected_value)

            return run_test

        def to_be_of_type(self, expected_type) -> Callable:
            def run_test(_=None):
                self.assertIsInstance(actual_value, expected_type)

            return run_test

        def to_equal_list(self, expected_list) -> Callable:
            def run_test(_=None):
                self.assertListEqual(actual_value, expected_list)

            return run_test

        def to_contain(self, expected_element) -> Callable:
            def run_test(_=None):
                self.assertTrue(expected_element in actual_value)

            return run_test

        def to_contain_all(self, expected_elements) -> Callable:
            def run_test(_=None):
                is_contained = True
                for element in expected_elements:
                    if element not in actual_value:
                        is_contained = False
                        break
                self.assertTrue(is_contained)

            return run_test

        def to_be_true(self) -> Callable:
            def run_test(_=None):
                self.assertTrue(actual_value)

            return run_test

        def to_be_false(self) -> Callable:
            def run_test(_=None):
                self.assertFalse(actual_value)

            return run_test

        def to_be_none(self) -> Callable:
            def run_test(_=None):
                self.assertIsNone(actual_value)

            return run_test

        def to_not_be_none(self) -> Callable:
            def run_test(_=None):
                self.assertIsNotNone(actual_value)

            return run_test

        def to_be_a_list(self) -> Callable:
            return self.to_be_of_type(list)

        def to_be_a_dict(self) -> Callable:
            return self.to_be_of_type(dict)

        def to_be_a_set(self) -> Callable:
            return self.to_be_of_type(set)

        def to_be_of_length(self, expected_length: int) -> Callable:
            def run_test(_=None):
                self.assertEqual(len(actual_value), expected_length)

            return run_test

    return Asserter()



"""
unittest-specs

A simple wrapper around the unittest package with different styles for specs
"""

__version__ = "1.1.1"
__author__ = "Phillip Goellner"


from unittest_specs.simple_test_spec import SimpleFlatSpec
from unittest_specs.fun_test_spec import describe, it, expect



import unittest

from unittest_specs.simple_test_spec import SimpleFlatSpec

parameterize = SimpleFlatSpec.parameterize
intercept = SimpleFlatSpec.intercept


def get_empty_test_case():
    def __(i):
        pass

    return type('', (unittest.TestCase,), {'subTest': __})


class SimpleTestDSL(unittest.TestCase):

    def test_param_decorator_should_execute_without_error(self):
        @parameterize(params=[])
        def empty_function(_):
            return 1

        empty_function(get_empty_test_case())

    @parameterize(params=[
        (1, 1),
        (2, 2),
    ])
    def test_should_decorate_class_member(self, left_param, right_param):
        self.assertEqual(left_param, right_param)

    def test_should_handle_exceptions_with_args_correctly(self):
        def exception_raiser(some_arg, another_arg=0):
            if some_arg == 1 and another_arg == 32:
                raise Exception()

        SimpleFlatSpec().expect(exception_raiser).to_raise(Exception, 1, another_arg=32)

    @parameterize(params=[
        ([1, 2, 3], 3),
        ('abcde', 'd'),
        ({0: 1}, 0),
        ((0, 1, 2, 3), 0),
    ])
    def test_should_detect_contained_correctly(self, given_input, expected_element):
        SimpleFlatSpec().expect(given_input).to_contain(expected_element)

    @parameterize(params=[
        ([1, 2, 3], [1, 3]),
        ({0: 1}, [0]),
        ((0, 1, 2, 3), [1, 3, 0]),
    ])
    def test_should_detect_contained_correctly(self, given_input, expected_elements):
        SimpleFlatSpec().expect(given_input).to_contain_all(expected_elements)


class SimpleAsserterTest(unittest.TestCase):
    def setUp(self) -> None:
        self.test_case = SimpleFlatSpec()

    def test_asserter_should_store_value_correctly(self):
        actual = self.test_case.expect(1)._actual_value
        self.assertEqual(1, actual)

    def test_asserter_should_detect_equality_correctly(self):
        self.test_case.expect('test').to_be('test')

    def test_asserter_should_detect_difference_correctly(self):
        self.test_case.expect('test').to_not_be('Test')

    def test_asserter_should_detect_type_correctly(self):
        self.test_case.expect('test').to_be_of_type(str)

    def test_asserter_should_test_list_equality_correctly(self):
        self.test_case.expect([1, 2, 3]).to_equal_list([1, 2, 3])

    def test_asserter_should_test_contains_correctly(self):
        self.test_case.expect([1, 2, 3]).to_contain(1)

    def test_asserter_should_test_contains_all_correctly(self):
        self.test_case.expect([1, 2, 3]).to_contain_all((2, 3, 1,))

    def test_asserter_should_test_true_correctly(self):
        self.test_case.expect(True).to_be_true()

    def test_asserter_should_test_false_correctly(self):
        self.test_case.expect(False).to_be_false()

    def test_asserter_should_test_none_correctly(self):
        self.test_case.expect(None).to_be_none()

    def test_asserter_should_test_not_none_correctly(self):
        self.test_case.expect('some').to_not_be_none()

    def test_asserter_should_recognize_set_correctly(self):
        self.test_case.expect({1, 2}).to_be_a_set()

    def test_asserter_should_recognize_list_correctly(self):
        self.test_case.expect([]).to_be_a_list()

    def test_asserter_should_recognize_dict_correctly(self):
        self.test_case.expect({}).to_be_a_dict()

    def test_asserter_should_check_length_correctly(self):
        self.test_case.expect({}).to_be_of_length(0)

    def test_should_handle_exceptions_correctly(self):
        def exception_raiser():
            raise Exception()

        self.test_case.expect(exception_raiser).to_raise(Exception)

    @intercept(Exception)
    def test_should_intercept_exceptions(self):
        def exception_raiser():
            raise Exception()

        exception_raiser()



from setuptools import setup

import unittest_specs

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="unittest-specs",
    version=unittest_specs.__version__,
    description="A simple wrapper around the unittest package with different styles for specs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phillipgoellner/unittest_specs",
    project_urls={
        "Source (GitHub)": "https://github.com/phillipgoellner/unittest_specs",
        "Source (GitLab)": "https://gitlab.com/goellnerphillip/unittest_specs"
    },
    author=unittest_specs.__author__,
    license="BSD Three-Clause License",
    packages=["unittest_specs"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
    ]
)
